from typing import NamedTuple, Self


class Padding(NamedTuple):
    top: int
    right: int
    bottom: int
    left: int

    @property
    def vertical(self) -> int:
        return self.top + self.bottom

    @property
    def horizontal(self) -> int:
        return self.left + self.right

    @classmethod
    def zero(cls) -> Self:
        return cls(0, 0, 0, 0)

    @classmethod
    def uniform(cls, value: int) -> Self:
        return cls(value, value, value, value)

    @classmethod
    def balanced(cls, top_bottom: int, left_right: int) -> Self:
        return cls(top_bottom, left_right, top_bottom, left_right)

    @classmethod
    def parse(cls, spec: str) -> Self:
        ps = [int(v.strip()) for v in spec.split(",")]
        if len(ps) == 1:
            return cls.uniform(ps[0])
        elif len(ps) == 2:
            return cls.balanced(ps[1], ps[0])
        elif len(ps) == 4:
            return cls(ps[0], ps[1], ps[2], ps[3])
        raise ValueError("expected p or px,py or pt,pr,pb,pl")



import curses
from dataclasses import dataclass, field
from typing import Self

from pyte.screens import Char

from minimux.colour import Colour, ColourManager, DefaultColour
from minimux.utils import combine


@dataclass
class Attr:
    fg: Colour = field(default_factory=lambda: DefaultColour())
    bg: Colour = field(default_factory=lambda: DefaultColour())
    blink: bool | None = None
    bold: bool | None = None
    dim: bool | None = None
    reverse: bool | None = None
    standout: bool | None = None
    underline: bool | None = None
    italic: bool | None = None

    @classmethod
    def from_char(cls, char: Char) -> Self:
        "Get the curses attribute of a Char from the pty"

        return cls(
            fg=Colour.new(char.fg),
            bg=Colour.new(char.bg),
            blink=char.blink or None,
            bold=char.bold or None,
            reverse=char.reverse or None,
            underline=char.underscore or None,
            italic=char.italics or None,
        )

    def __or__(self, other: object) -> "Attr":
        """
        Combine two attributes, with any values set in other
        overriding the values in self
        """
        if not isinstance(other, Attr):
            raise TypeError

        return Attr(
            fg=self.fg if other.fg == DefaultColour() else other.fg,
            bg=self.bg if other.bg == DefaultColour() else other.bg,
            blink=combine(self.blink, other.blink),
            bold=combine(self.bold, other.bold),
            dim=combine(self.dim, other.dim),
            reverse=combine(self.reverse, other.reverse),
            standout=combine(self.standout, other.standout),
            underline=combine(self.underline, other.underline),
            italic=combine(self.italic, other.italic),
        )

    def __call__(self, colour_manager: ColourManager | None = None) -> int:
        """
        Convert the attribute into a curses attribute value
        """
        attr = 0
        if colour_manager is not None:
            pair = colour_manager.make_pair(self.fg, self.bg)
            attr = curses.color_pair(pair)
        if self.blink:
            attr |= curses.A_BLINK
        if self.bold:
            attr |= curses.A_BOLD
        if self.dim:
            attr |= curses.A_DIM
        if self.reverse:
            attr |= curses.A_REVERSE
        if self.standout:
            attr |= curses.A_STANDOUT
        if self.underline:
            attr |= curses.A_UNDERLINE
        if self.italic:
            attr |= curses.A_ITALIC
        return attr



from typing import Generator

import pyte
import pyte.modes
from pyte.screens import Char

BUFFER_MAX_SIZE = 10000


class Buffer:
    """
    Holds lines of output of a command. Only enough lines to fill the
    height of the output space (maxrows) are kept, and lines are split
    so that each line is at most maxcols in length
    """

    def __init__(
        self,
        maxcols: int,
        maxrows: int,
        tabsize: int,
    ):
        self.buf: str = ""
        self.tabsize = tabsize

        # virtual terminal
        self.screen = pyte.Screen(maxcols, maxrows)
        self.stream = pyte.Stream()
        self.stream.attach(self.screen)
        self.reset()

    @property
    def rows(self) -> int:
        return self.screen.lines

    @property
    def cols(self) -> int:
        return self.screen.columns

    def push(self, data: str):
        "appends data to the buffer"
        self.buf += data
        if len(self.buf) > BUFFER_MAX_SIZE:
            self.buf = self.buf[len(self.buf) - BUFFER_MAX_SIZE :]
        self.stream.feed(data)

    def reset(self, clear_buffer: bool = True):
        "clears the buffer of all data"
        if clear_buffer:
            self.buf = ""
        self.screen.reset()
        self.screen.mode.add(pyte.modes.LNM)
        ntabs = self.screen.columns // self.tabsize
        self.screen.tabstops = set(i * self.tabsize for i in range(ntabs))

    def resize(
        self,
        *,
        maxcols: int | None = None,
        maxrows: int | None = None,
    ):
        "resizes the buffer"
        self.screen.resize(maxrows, maxcols)
        self.reset(False)
        self.stream.feed(self.buf)

    def lines(self) -> Generator[list[Char], None, None]:
        for row in range(self.screen.lines):
            bufline = self.screen.buffer[row]
            yield [bufline[col] for col in range(self.screen.columns)]



import configparser
import shlex
from dataclasses import dataclass
from pathlib import Path
from typing import TextIO

import minimux.utils as utils
from minimux.attr import Attr
from minimux.colour import Colour, DefaultColour
from minimux.padding import Padding


@dataclass
class Config:
    title: str | None
    content: "Config.Element"
    sep_attr: Attr
    title_attr: Attr

    @dataclass
    class Element:
        name: str
        attr: Attr
        weight: int

    @dataclass
    class Command(Element):
        command: list[str]
        title: str | None
        input: str | None
        padding: Padding
        title_attr: Attr
        no_close_stdin: bool
        charset: str
        cwd: Path
        watch: int
        tabsize: int

    @dataclass
    class Panel(Element):
        vertical: bool
        children: "list[Config.Element]"

    @classmethod
    def from_parser(cls, parser: "MiniMuxConfigParser") -> "Config":
        main = parser["main"]
        title = main.pop("title", None)
        content = parser.create_panels(main, "", Attr())
        base_attr = parser.parse_attrs(main)
        sep_attrs = base_attr
        if "seperator" in parser:
            sep_attrs = base_attr | parser.parse_attrs(parser["seperator"])
        title_attrs = base_attr
        if "title" in parser:
            title_attrs = base_attr | parser.parse_attrs(parser["title"])

        return cls(title, content, sep_attrs, title_attrs)

    @classmethod
    def from_file(cls, f: TextIO, cwd: Path) -> "Config":
        parser = MiniMuxConfigParser(cwd)
        parser.read_file(f)
        return cls.from_parser(parser)


class MiniMuxConfigParser(configparser.ConfigParser):
    def __init__(self, cwd: Path):
        super().__init__(
            delimiters=["="],
            comment_prefixes=["#"],
            default_section="default",
            converters={
                "list": utils.split_list,
                "path": Path,
                "colour": Colour.new,
            },
        )
        self.cwd = cwd

    def create_panels(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Config.Element:
        if "command" in section:
            return self.parse_command(section, prefix, default_attr)
        elif "panels" in section:
            return self.parse_panel(section, prefix, default_attr)
        else:
            raise ValueError("command or panels must be specified")

    def parse_command(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Config.Command:
        title = section.get("title", None)
        shell = section.getboolean("shell", False)
        if shell:
            command = ["sh", "-c", section["command"]]
        else:
            command = shlex.split(section["command"])
        attr = default_attr | self.parse_attrs(section)
        title_attr = attr
        if s := section.get("title_attr", None):
            title_attr = attr | self.parse_attrs(self[s])
        weight = section.getint("weight", 1)
        input = section.get("input", None)
        padding = Padding.parse(section.get("padding", "0"))
        no_close_stdin = section.getboolean("no_close_stdin", False)
        charset = section.get("charset", "utf-8")
        cwd: Path = section.getpath("cwd", self.cwd)
        watch = section.getint("watch", 0)
        tabsize = section.getint("tabsize", 4)

        return Config.Command(
            prefix + ":" + section.name,
            attr,
            weight,
            command,
            title,
            input,
            padding,
            title_attr,
            no_close_stdin,
            charset,
            cwd,
            watch,
            tabsize,
        )

    def parse_panel(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Config.Panel:
        prefix += ":" + section.name
        vertical = section.getboolean("vertical", False)
        attr = default_attr | self.parse_attrs(section)
        children: list[Config.Element] = []
        weight = section.getint("weight", 1)
        for i, subsection in enumerate(section.getlist("panels", [])):
            child = self.create_panels(
                self[subsection],
                f"{prefix}.{i}",
                attr,
            )
            children.append(child)

        return Config.Panel(prefix, attr, weight, vertical, children)

    def parse_attrs(self, section: "configparser.SectionProxy") -> Attr:
        attr = Attr()
        attr.fg = section.getcolour("fg", DefaultColour())
        attr.bg = section.getcolour("bg", DefaultColour())
        attr.blink = section.getboolean("blink", None)
        attr.bold = section.getboolean("bold", None)
        attr.dim = section.getboolean("dim", None)
        attr.reverse = section.getboolean("reverse", None)
        attr.standout = section.getboolean("standout", None)
        attr.underline = section.getboolean("underline", None)
        attr.italic = section.getboolean("italic", None)
        return attr



import curses
import re
from dataclasses import dataclass
from typing import Self

from minimux.utils import Counter, UniqueDict

re_hexcolour = re.compile(r"#?([A-Fa-f0-9]{1,2})([A-Fa-f0-9]{1,2})([A-Fa-f0-9]{1,2})")
re_rgbcolour = re.compile(r"rgb\s*\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)")


@dataclass(frozen=True)
class Colour:
    @staticmethod
    def new(spec: str) -> "Colour":
        if spec == "default":
            return DefaultColour()
        elif m := re_hexcolour.fullmatch(spec):
            return RGBColour.from_rgb(
                int(m[1], 16),
                int(m[2], 16),
                int(m[3], 16),
            )
        elif m := re_rgbcolour.fullmatch(spec):
            return RGBColour.from_rgb(
                int(m[1]),
                int(m[2]),
                int(m[3]),
            )
        return NamedColour(spec)


@dataclass(frozen=True)
class NamedColour(Colour):
    name: str


@dataclass(frozen=True)
class RGBColour(Colour):
    r: int
    g: int
    b: int

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int) -> Self:
        return cls(r * 1000 // 256, g * 1000 // 256, b * 1000 // 256)


@dataclass(frozen=True)
class DefaultColour(Colour):
    pass


@dataclass(frozen=True)
class ColourPair:
    fg: Colour
    bg: Colour


class ColourManager:
    def __init__(self):
        self.colours: UniqueDict[Colour, int] = UniqueDict()
        self.next_colour = Counter(8, curses.COLORS)
        self.colour_pairs: UniqueDict[ColourPair, int] = UniqueDict()
        self.next_colour_pair = Counter(1, curses.COLOR_PAIRS)

        # default colour
        self.colours[DefaultColour()] = -1

        # inbuilt curses colours
        self.colours[NamedColour("black")] = curses.COLOR_BLACK
        self.colours[NamedColour("blue")] = curses.COLOR_BLUE
        self.colours[NamedColour("cyan")] = curses.COLOR_CYAN
        self.colours[NamedColour("green")] = curses.COLOR_GREEN
        self.colours[NamedColour("magenta")] = curses.COLOR_MAGENTA
        self.colours[NamedColour("red")] = curses.COLOR_RED
        self.colours[NamedColour("white")] = curses.COLOR_WHITE
        self.colours[NamedColour("yellow")] = curses.COLOR_YELLOW

        # terminal "bright" colours
        self._init("brightblack", 85, 87, 83)
        self._init("brightblue", 124, 158, 203)
        self._init("brightcyan", 112, 223, 224)
        self._init("brightgreen", 158, 224, 85)
        self._init("brightmagenta", 166, 129, 165)
        self._init("brightred", 220, 63, 54)
        self._init("brightwhite", 238, 238, 236)
        self._init("brightyellow", 249, 234, 107)

        # ensure we don't overwrite the named colours
        self.next_colour.clamp()

    def make_pair(self, fg: Colour, bg: Colour) -> int:
        "Returns a curses colour pair id for the given colours"
        pair = ColourPair(fg, bg)

        if pair in self.colour_pairs:
            return self.colour_pairs[pair]

        res = self.next_colour_pair()
        curses.init_pair(res, self.make_colour(fg), self.make_colour(bg))
        self.colour_pairs[pair] = res
        return res

    def make_colour(self, colour: Colour) -> int:
        "Returns a curses colour id for the given colour"
        if colour in self.colours:
            return self.colours[colour]

        # create a new colour if it is an rgb definition
        if isinstance(colour, RGBColour):
            res = self.next_colour()
            curses.init_color(res, colour.r, colour.g, colour.b)
            self.colours[colour] = res
            return res

        # invalid color, return default
        return -1

    def _init(self, name: str, r: int, g: int, b: int) -> int:
        res = self.next_colour()
        colour = RGBColour.from_rgb(r, g, b)
        curses.init_color(res, colour.r, colour.g, colour.b)
        self.colours[NamedColour(name)] = res
        return res



import curses
from pathlib import Path

import click

from minimux import MiniMux, __version__
from minimux.config import Config


def _main(config_file: Path | None, directory: Path | None):
    # default to current directory
    if directory is None:
        directory = Path(".")

    # attempt to locate minimux.ini, first in current directory then
    # in directory
    if config_file is None:
        if Path("./minimux.ini").exists():
            config_file = Path("./minimux.ini")
        elif (directory / "minimux.ini").exists():
            config_file = directory / "minimux.ini"
        else:
            raise ValueError("could not locate a minimux.ini file")

    # load config
    with open(config_file) as f:
        config = Config.from_file(f, directory)

    # run
    def wrapper(stdscr: "curses.window"):
        minimux = MiniMux(stdscr, config, directory)
        minimux.run()

    curses.wrapper(wrapper)


@click.command("minimux")
@click.version_option(
    __version__,
    "--version",
    "-v",
)
@click.option(
    "--directory",
    "-d",
    type=click.Path(
        exists=True,
        file_okay=False,
        path_type=Path,
    ),
    default=None,
    help="The directory to set as the cwd for commands. Additionally, if CONFIG_FILE is not set and there is no minimux.ini file in the current directory, this directory will be searched for a minimux.ini file to use",
)
@click.option(
    "--debug",
    "-g",
    is_flag=True,
    help="Display a full traceback when an exception is thrown",
)
@click.argument(
    "config_file",
    type=click.Path(
        exists=True,
        dir_okay=False,
        path_type=Path,
    ),
    required=False,
    default=None,
)
def main(config_file: Path | None, directory: Path | None, debug: bool):
    try:
        _main(config_file, directory)
    except Exception as e:
        if debug:
            raise
        exit("minimux: fatal: " + str(e))
    except:
        if debug:
            raise
        exit("minimux: fatal: an unhandled exception occurred")


if __name__ == "__main__":
    main()



import curses
import threading
from pathlib import Path

from minimux.colour import ColourManager
from minimux.config import Config
from minimux.range import Range, Range2d
from minimux.runner import Runner
from minimux.utils import char_eq

__version__ = "2.0.0"
__version_info__ = (2, 0, 0)
__author__ = "Dominic Price"


class MiniMux:
    def __init__(
        self,
        stdscr: "curses.window",
        config: Config,
        cwd: Path,
    ):
        self.stdscr = stdscr
        self.config = config
        self.lock = threading.Lock()
        self.cwd = cwd
        self.runners: dict[str, Runner] = {}
        self.colour_manager = ColourManager()

    def run(self):
        curses.curs_set(False)
        curses.use_default_colors()

        self.runners = self.get_runners(self.config.content)
        self.joins = [runner.start() for runner in self.runners.values()]
        self.create_windows()

        try:
            # handle user input
            while True:
                if self.stdscr.getch() == curses.KEY_RESIZE:
                    self.create_windows()
        except KeyboardInterrupt:
            # keyboard interrupts are a normal way to exit
            pass
        finally:
            # kill all runners and wait for them to terminate
            # before exiting curses mode to avoid leaving the
            # terminal in a bad state
            rows, cols = self.stdscr.getmaxyx()
            for runner in self.runners.values():
                runner.stop()
            self.stdscr.clear()
            self.stdscr.move(rows // 2, 0)
            self.stdscr.addstr("Terminating...".center(cols))
            self.stdscr.refresh()

            # join all threads, collecting errors
            errors = []
            for join in self.joins:
                if err := join():
                    errors.append(err)

            # print errors and exit with non-zero status code
            if len(errors) > 0:
                for err in errors:
                    print(err)
                exit(1)

    def get_runners(self, content: Config.Element) -> dict[str, Runner]:
        if isinstance(content, Config.Panel):
            res: dict[str, Runner] = {}
            for child in content.children:
                res.update(self.get_runners(child))
            return res
        elif isinstance(content, Config.Command):
            return {content.name: Runner(content, self.doupdate, self.colour_manager)}
        else:
            raise TypeError

    def doupdate(self):
        """
        Call curses.doupdate() protected by a mutex so that it is
        safe to call from multiple threads
        """
        with self.lock:
            curses.doupdate()

    def create_windows(self):
        rows, cols = self.stdscr.getmaxyx()
        self.stdscr.clear()

        start_row = 0
        if self.config.title:
            self.stdscr.move(0, 0)
            self.stdscr.addstr(
                self.config.title.center(cols),
                self.config.title_attr(self.colour_manager),
            )
            self.hsep(1, Range(0, cols))
            start_row = 2

        self.init_content(
            self.config.content,
            Range(start_row, rows),
            Range(0, cols),
        )
        self.stdscr.refresh()

    def init_content(
        self,
        content: Config.Element,
        range_y: Range,
        range_x: Range,
    ):
        """Recursively draw the static components for content and
        initialise the runners for any commands"""
        if isinstance(content, Config.Panel):
            self.init_panel(content, range_y, range_x)
        elif isinstance(content, Config.Command):
            self.init_command(content, range_y, range_x)
        else:
            raise TypeError

    def init_panel(
        self,
        panel: Config.Panel,
        range_y: Range,
        range_x: Range,
    ):
        """Recursively draw the static components for a panel and
        initialise the runners from any commands"""
        if len(panel.children) == 0:
            return

        if panel.vertical:
            o = range_y.start
            subh = (range_y.length) // sum(c.weight for c in panel.children)
            i = 0
            for child in panel.children:
                subrange_y = Range.offset(i * subh, (i + child.weight) * subh, o)
                if i == len(panel.children) - 1:
                    subrange_y = Range(subrange_y.start, range_y.end)
                if i != 0:
                    self.hsep(subrange_y.start, range_x)
                    subrange_y = subrange_y.trim(top=1)
                i += child.weight
                self.init_content(child, subrange_y, range_x)
        else:
            i = 0
            o = range_x.start
            subw = (range_x.length) // sum(c.weight for c in panel.children)
            for child in panel.children:
                subrange_x = Range.offset(i * subw, (i + child.weight) * subw, o)
                subrange_y = range_y
                if i == len(panel.children) - 1:
                    subrange_x = Range(subrange_x.start, range_x.end)
                if i != 0:
                    self.vsep(subrange_x.start, range_y)
                    subrange_x = subrange_x.trim(top=1)
                self.init_content(child, subrange_y, subrange_x)
                i += child.weight

    def init_command(
        self,
        command: Config.Command,
        range_y: Range,
        range_x: Range,
    ):
        """Draw the static components for a command initialise the
        associated runner"""
        if command.title is not None:
            self.stdscr.move(range_y.start, range_x.start)
            self.stdscr.addstr(
                " " * (range_x.length),
                command.attr(self.colour_manager),
            )
            self.center(
                command.title,
                range_y.start,
                range_x,
                command.title_attr(self.colour_manager),
            )
            range_y = range_y.trim(top=1)
        self.runners[command.name].init(self.stdscr, Range2d(range_y, range_x))

    def center(self, s: str, y: int, x_range: Range, attr: int):
        pad_left = (x_range.length - len(s)) // 2
        self.stdscr.move(y, x_range.start + pad_left)
        self.stdscr.addstr(s, attr)

    def hsep(self, y: int, x_range: Range):
        """Draw a horizontal seperator line, combining with existing
        separators to form tees and crosses"""
        attr = self.config.sep_attr(self.colour_manager)
        _, max_x = self.stdscr.getmaxyx()
        if x_range.start > 0:
            if char_eq(self.stdscr.inch(y, x_range.start - 1), curses.ACS_SBSB):
                self.stdscr.move(y, x_range.start - 1)
                self.stdscr.addch(curses.ACS_SBSB, attr)
        if x_range.end <= max_x:
            if char_eq(self.stdscr.inch(y, x_range.end), curses.ACS_SBSB):
                self.stdscr.move(y, x_range.end)
                self.stdscr.addch(curses.ACS_SBSB, attr)
        for i in x_range.values():
            self.stdscr.move(y, i)
            self.stdscr.addch(curses.ACS_BSBS, attr)
        self.stdscr.noutrefresh()

    def vsep(self, x: int, y_range: Range):
        """Draw a vertical seperator line, combining with existing
        separators to form tees and crosses"""
        attr = self.config.sep_attr(self.colour_manager)
        max_y, _ = self.stdscr.getmaxyx()
        if y_range.start > 0:
            if char_eq(self.stdscr.inch(y_range.start - 1, x), curses.ACS_BSBS):
                self.stdscr.move(y_range.start - 1, x)
                self.stdscr.addch(curses.ACS_BSSS, attr)
        if y_range.end < max_y:
            if char_eq(self.stdscr.inch(y_range.end, x), curses.ACS_BSBS):
                self.stdscr.move(y_range.end, x)
                self.stdscr.addch(curses.ACS_BSSS, attr)
        for i in y_range.values():
            self.stdscr.move(i, x)
            self.stdscr.addch(curses.ACS_SBSB, attr)
        self.stdscr.noutrefresh()



import atexit
import curses
import os
import subprocess
import threading
import time
from typing import Callable

from minimux.attr import Attr
from minimux.buffer import Buffer
from minimux.colour import ColourManager
from minimux.config import Config
from minimux.range import Range2d

MAX_READ_BUFFER = 1024


class Runner:
    def __init__(
        self,
        command: Config.Command,
        doupdate: Callable[[], None],
        colour_manager: ColourManager,
    ):
        self.command = command
        self.doupdate = doupdate
        self.win: "curses.window | None" = None
        self.colour_manager = colour_manager
        self.buf = Buffer(0, 0, command.tabsize)
        self._stopped = threading.Event()
        self.err: Exception | None = None
        self.lock = threading.Lock()

    def init(self, stdscr: "curses.window", bounds: Range2d):
        "Set the curses window for the runner"
        with self.lock:
            if self.win is not None:
                del self.win
            self.win = stdscr.subwin(
                bounds.rows.length,
                bounds.cols.length,
                bounds.rows.start,
                bounds.cols.start,
            )
            bkgd = self.command.attr(self.colour_manager)
            self.win.bkgdset(" ", bkgd)
            self.win.bkgd(" ", bkgd)
            self.buf.resize(
                maxrows=bounds.rows.length - self.command.padding.vertical,
                maxcols=bounds.cols.length - self.command.padding.horizontal,
            )
        self._flush()

    def start(self) -> Callable[[], Exception | None]:
        "Start the runner in a background thread, returns immediately"

        def run_safe():
            try:
                self.run()
            except Exception as e:
                self.err = e
            except:
                self.err = RuntimeError("unhandled error")

        t = threading.Thread(target=run_safe)
        t.start()

        def join():
            t.join(5)
            return self.err

        return join

    def stop(self):
        self._stopped.set()

    def run(self):
        "Start the runner until the process exits"
        self._stopped.clear()
        next_launch = time.time()

        while not self._stopped.wait(next_launch - time.time()):
            self._clear()
            next_launch = time.time() + self.command.watch

            # spawn the process
            proc = subprocess.Popen(
                self.command.command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                stdin=subprocess.PIPE,
                text=True,
                cwd=self.command.cwd,
            )

            def terminate():
                if proc.stdout:
                    proc.stdout.close()
                if proc.stdin:
                    proc.stdin.close()
                proc.terminate()

            # ensure process is terminated at program exit
            atexit.register(terminate)

            # if we are watching, ensure it is terminated when we need
            # the loop to next execute
            timer: threading.Timer | None = None
            if self.command.watch > 0:
                timer = threading.Timer(self.command.watch, terminate)
                timer.start()

            # poll for process output
            return_code = self._poll_process(proc)

            # cancel the timer
            if timer is not None:
                timer.cancel()

            # unregister terminate function at program exit
            atexit.unregister(terminate)

            # stop loop if we are not watching the output
            if self.command.watch <= 0:
                self._write(f"** Exited with return code {return_code} **")
                return

    def _poll_process(self, proc: subprocess.Popen[str]) -> int:
        if proc.stdin is None:
            self._write("error: could not grab a handle to stdin\n")
            return -1
        if proc.stdout is None:
            self._write("error: could not grab a handle to stdout\n")
            return -1

        # send input to command's stdin
        if self.command.input is not None:
            proc.stdin.write(self.command.input + "\n")
            proc.stdin.flush()

        # close stdin unless instructed not to
        if not self.command.no_close_stdin:
            proc.stdin.close()

        # poll process for output
        fileno = proc.stdout.fileno()
        while line := os.read(fileno, MAX_READ_BUFFER):
            self._write(line)

        # ensure process is terminated and write output code to buffer
        return proc.wait()

    def _write(self, data: str | bytes):
        if isinstance(data, bytes):
            data = data.decode(self.command.charset, errors="replace")
        with self.lock:
            self.buf.push(data)
        self._flush()

    def _clear(self):
        with self.lock:
            self.buf.reset()

    def _flush(self):
        with self.lock:
            if self.win is None:
                return
            self.win.clear()
            for row, line in enumerate(self.buf.lines()):
                for col, char in enumerate(line):
                    attr = self.command.attr | Attr.from_char(char)
                    self.win.move(
                        self.command.padding.top + row,
                        self.command.padding.left + col,
                    )
                    try:
                        self.win.addch(char.data, attr(self.colour_manager))
                    except curses.error:
                        # trying to write to the bottom-right corner
                        # raises an error after the character is added
                        # https://docs.python.org/3/library/curses.html#curses.window.addch
                        if row == self.buf.rows - 1 and col == self.buf.cols - 1:
                            pass
                        else:
                            raise
            self.win.noutrefresh()
        self.doupdate()



from typing import Generator, NamedTuple


class Range(NamedTuple):
    start: int
    end: int

    @classmethod
    def offset(cls, start: int, end: int, offset: int) -> "Range":
        return cls(start + offset, end + offset)

    @property
    def length(self) -> int:
        return self.end - self.start

    def trim(self, *, top: int = 0, tail: int = 0) -> "Range":
        return Range(self.start + top, self.end - tail)

    def values(self) -> Generator[int, None, None]:
        yield from range(self.start, self.end)


class Range2d(NamedTuple):
    rows: Range
    cols: Range



import curses
from typing import Generic, TypeVar

T = TypeVar("T")


def combine(t1: T, t2: T | None) -> T:
    "return t1 if t2 is None, else t2"
    if t2 is None:
        return t1
    return t2


def char_eq(a: int, b: int) -> bool:
    "returns True if a and b are equal characters, ignoring attributes"
    return (a & curses.A_CHARTEXT) == (b & curses.A_CHARTEXT)


def split_list(v: str) -> list[str]:
    "converts a comma separated string into a list, omitting empty elements"
    res: list[str] = []
    for elem in v.split(","):
        e = elem.strip()
        if e:
            res.append(e)
    return res


class Counter:
    def __init__(self, min_value: int, max_value: int):
        self.value = min_value
        self.min_value = min_value
        self.max_value = max_value

    def __call__(self) -> int:
        res = self.value
        self.value += 1
        if self.value > self.max_value:
            self.value = self.min_value
        return res

    def clamp(self):
        self.min_value = self.value


K = TypeVar("K")
V = TypeVar("V")


class UniqueDict(Generic[K, V]):
    """
    UniqueDict ensures that each value is unique, removing old keys
    if a new key with the same value is added
    """

    def __init__(self):
        self.forward: dict[K, V] = {}
        self.backward: dict[V, K] = {}

    def __setitem__(self, key: K, value: V):
        if key in self.forward:
            del self.backward[self.forward[key]]
        self.forward[key] = value
        if value in self.backward:
            del self.forward[self.backward[value]]
        self.backward[value] = key

    def __getitem__(self, key: K) -> V:
        return self.forward[key]

    def __contains__(self, key: K) -> bool:
        return key in self.forward

    def __len__(self) -> int:
        return len(self.forward)
